// =====================================================================================================================
// bmcremedychange.js
//
// Main file for the BMC Remedy change integration service.
//
// =====================================================================================================================
importPackage(java.util);
importPackage(java.io);
importPackage(java.text);

importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessageImpl);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.alarmpoint.integrationagent.security.EncryptionUtils);
importClass(Packages.com.alarmpoint.integrationagent.soap.exception.SOAPRequestException);

importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);

importClass(Packages.org.apache.commons.httpclient.Header);
importClass(Packages.org.apache.commons.httpclient.HttpVersion);
importClass(Packages.org.mule.providers.http.HttpResponse);


// Core Javascript files provided by the IA as of 4.1 patch 5
load("integrationservices/bmcremedy/lib/javascript/core/baseclass.js");
load("integrationservices/bmcremedy/lib/javascript/core/logger.js");
load("integrationservices/bmcremedy/lib/javascript/webservices/wsutil.js");
load("integrationservices/bmcremedy/lib/javascript/webservices/soapfault.js");
load("integrationservices/bmcremedy/lib/javascript/xmatters/xmattersws.js");
load("integrationservices/bmcremedy/util.js");

// Integration-specific Javascript files
load("integrationservices/bmcremedy/util.js")
load("integrationservices/bmcremedy/bmcremedychange/configuration.js");
load("integrationservices/bmcremedy/bmcremedychange/xm-chg-infrastructure-change-ws.js")
load("integrationservices/bmcremedy/bmcremedychange/xm-chg-change-interface-ws.js")
load("integrationservices/bmcremedy/bmcremedychange/xm-chg-change-ap-detail-signature-ws.js")
load("integrationservices/bmcremedy/bmcremedychange/updateChangeRequest.js")


var xStream = new XStream(new PureJavaReflectionProvider());
var exceptionString = "JavaException: ";

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * apia_http
 *
 * This is the entry point for HTTP requests from Remedy that initiate data load operations.
 *
 * This method expects that the POST body contains a SOAP 1.1 message indicating the data load action to be taken.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function apia_http(httpRequestProperties, httpResponse)
{
    log.info("Enter - apia_http");
    log.debug("httpRequestProperties: " + httpRequestProperties);

    var httpRequest = null;
    var responseBuilder = null;
    var responseBody = null;

    try
    {
        httpRequest = httpRequestProperties.getProperty("http.request");
//      log.debug("Handling request to: " + httpRequest);

        processRequestBody(new XML(httpRequestProperties.getProperty("REQUEST_BODY")));

        responseBody = makeSoapResponseBody(APIA_HTTP_RECEIVED, "Request was submitted for processing");
    }
    catch (e)
    {
        log.warn("apia_http: caught Exception - name: [" + e.name + "], message [" + e.message + "]");
        responseBody = makeSoapResponseBody(APIA_HTTP_ERROR, e.toString());
    }
    finally
    {
        httpResponse.setStatusLine(HttpVersion.HTTP_1_1, 200);
        httpResponse.setHeader(new Header("Content-Type", httpRequestProperties.getProperty("Content-Type")));
        httpResponse.setBodyString(responseBody);
    }

    log.debug("apia_http - response body [" + responseBody + "]");
    log.info("Exit - apia_http");

    return httpResponse;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * processRequestBody
 *
 * @param requestBody - the body of the HTTP Request
 * ---------------------------------------------------------------------------------------------------------------------
 */
function processRequestBody(requestBody)
{
    log.debug("Enter - processRequestBody");
    log.info("processRequestBody: request body [" + requestBody + "]");

    var request = new Object();

    request.action = String(requestBody.*::Body.*::TriggerRequest.*::action);
    request.id = String(requestBody.*::Body.*::TriggerRequest.*::id);
    request.requestType = String(requestBody.*::Body.*::TriggerRequest.*::requestType);

    if (isEmpty(request.action))
    {
        throw "Request action not found."
    }

    if (isEmpty(request.id))
    {
        throw "Request id not found."
    }

    if (isEmpty(request.requestType))
    {
        throw "Request type not found."
    }

    log.debug("Exit - processRequestBody");
    return processRequest(request);
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * processRequest
 *
 * Process the request from BMC Remedy, which has been unmarshalled into a object with properties for the:
 *
 * - request action (Delete or something else)
 * - request type (change or approval)
 * - id of the change or approval that the request applies to
 *
 * @param request -
 * ---------------------------------------------------------------------------------------------------------------------
 */
function processRequest(request)
{
    log.debug("Enter - processRequest");
    log.debug("processRequestBody: action [" + request.action + "], requestType [" + request.requestType + "], id [" + request.id + "]");

    var xmWebService = new xMattersWS();

    if (equalsIgnoreCase(request.action, REQUEST_ACTION_DELETE))
    {
        // Remedy can send requests to delete a change or approval request

        log.info("processRequest - deleting change or approval request [" + request.id + "]");
        xmWebService.sendDelAPXML(request.id);
    }
    else
    {
        // Request is a new or updated change or approval request

        var changeService = new XMCHGInfrastructureChangeWS(new WSUtil(), XM_CHG_INFRASTRUCTURE_CHANGE_WS_URL);
        var changeRequestId = "";

        // Even if we are processing an approval request notification, we need the ID of the Change Request.

        switch (String(request.requestType).toLowerCase())
        {
            case REQUEST_TYPE_CHANGE:
                changeRequestId = request.id;
                break;

            case REQUEST_TYPE_APPROVAL:
                // Retrieve the full description of the approval that the event is associated with, to get the Change Request
                var changeAPDetailSignatureService = new XMCHGChangeAPDetailSignatureWS(new WSUtil(), XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_URL);
                var qualification = XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_GET_APPROVAL_BY_ID_QUALIFICATION.replace("${REQUEST_ID}", String(request.id));

                var approvalList = changeAPDetailSignatureService.getApprovals(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, qualification);

                log.debug("processApproval - approvalList [" + approvalList + "]");

                changeRequestId = approvalList[0].Infrastructure_Change_ID;
                break;

            default:
                throw ("Unknown request type [" + request.requestType + "]");
        }

        // Retrieve the Remedy Change Request associated with the event
        var changeRequest = changeService.getChangeRequest(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, changeRequestId);

        if (changeRequest == null)
        {
            throw ("changeRequest with id [" + request.id + "] not found in Remedy");
        }

        log.debug("processApproval - changeRequest [" + changeRequest + "]");

        // Make the APXML message to be sent to xMatters using the fields in the Change Request
        var apxml = makeApxmlFrom(changeRequest, request.id, String(request.requestType).toLowerCase());

        // If this is notification of an approval, add the tokens for the approval object
        if (String(request.requestType).toLowerCase() == REQUEST_TYPE_APPROVAL)
        {
            xmWebService.addEventTokensFromObject(apxml, approvalList[0]);
            apxml.setToken("apapprovers-list-item-delimiter", ";");

            // Replace the change request id with the change detail signature request id as the incident_id
            apxml.setToken("request_id", request.id);
        }

        // Send the message to xMatters
        if (xmWebService.submitAPXML(apxml, DELETE_EXISTING_EVENTS))
        {
            // If submitAPXML() actually sent the message and didn't deduplicate it, add an annotation to the changeRequest.
            var changeInterfaceService = new XMCHGChangeInterfaceWS(new WSUtil(), XM_CHG_CHANGE_INTERFACE_WS_URL);

            setChangeWorkLog(changeRequest, true, notePrefix + request.requestType + " event forwarded to xMatters", "");
            changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, changeRequest);
        }
    }

    log.debug("Exit - processRequest");
    return apxml;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * makeApxmlFrom
 *
 * Constructs and returns an APXML message based on the information in the changeRequest and approvalList objects.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function makeApxmlFrom(changeRequest, id, requestType)
{
    log.debug("Enter - makeApxmlFrom");
    log.debug("makeApxmlFrom - changeRequest [" + changeRequest + "], id [" + id + "], requestType [" + requestType + "]");

    var apxml = ServiceAPI.createAPXML();
    apxml.setMethod("ADD");
    apxml.setSubclass("action");

    // Set up the APXML tokens that have the same names and values as the Remedy changeRequest properties
    new xMattersWS().addEventTokensFromObject(apxml, changeRequest);

    apxml.setToken("incident_id", id);
    apxml.setToken("requestType", requestType);

    // Add the xm_priority event token to set the priority of the event in xMatters (for 5.0)
    var xm_priority;

    switch (String(changeRequest.Priority).toLowerCase())
    {
        case "critical" :
            xm_priority = "HIGH";
            break;
        case "high" :
            xm_priority = "MEDIUM";
            break;
        default :
            xm_priority = "LOW";
    }

    apxml.setToken("xm_priority", xm_priority);

    log.debug("Exit - makeApxmlFrom");
    return apxml;
}

/**
 * New lifecycle method not used in this integration, having this here prevents a nuisance stacktrace
 */
function apia_webservice_init(){}

/**
 * =====================================================================================================================
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a
 * Web Service response, depending on the mechanism used to submit the APXML.
 * =====================================================================================================================
 */
function apia_response(apxml)
{
    log.debug("Enter - apia_response");
    log.debug("Method =[" + apxml.getMethod() + "]");

    var result = null;

    switch (String(apxml.getMethod()).toUpperCase())
    {
        case String(APXMLMessage.OK_METHOD).toUpperCase():
            result = handleOK(apxml);
            break;

        case String(APXMLMessage.ERROR_METHOD).toUpperCase():
            result = handleERROR(apxml);
            break;

        case String(APXMLMessage.SEND_METHOD).toUpperCase():
            result = handleSend(apxml);
            break;

        case String(APXMLMessage.REQUEST_METHOD).toUpperCase():
            result = handleRequest(apxml);
            break;

        default:
            throw("Unrecognized APXML method: " + apxml.getMethod());
    }

    log.debug("Exit - apia_response");
    return result;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleOK
 *
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleOK(apxml)
{
    var id = apxml.getTransactionId();
    log.info("APXML submission with transaction id of " + id + " succeeded.");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleERROR
 *
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleERROR(apxml)
{
    var id = apxml.getTransactionId();
    var code = apxml.getValue(APXMLMessage.ERROR_CODE);

    log.error("APXML submission with transaction id of " + id + " failed with error code " + code);
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * Sends are ESMs to the IA.
 * 
 * Messages sent as ESMs are typically device delivery annotation requests.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleSend(apxml)
{
    log.debug("Enter - handleSend");
    log.info("handleSend - apxml: [" + apxml + "]");

    handleResponse(apxml);

    log.debug("Exit - handleSend");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * Requests are ESRs to the IA.
 *
 * Messages sent as ESRs are typically user responses.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleRequest(apxml)
{
    log.debug("Enter - handleRequest");
    log.info("handleRequest - apxml: [" + apxml + "]");

    var xmWebService = new xMattersWS();
    var responseToXM = xmWebService.initializeResponse(apxml);

    try
    {
        handleResponse(apxml);
        responseToXM.setToken(xmWebService.TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    }
    catch (e)
    {
        var errMsg = "Caught exception: " + e;
        log.error(errMsg);
        responseToXM.setToken(xmWebService.TOKEN_NAME, "Exception", APXMLToken.Type.STRING);
        responseToXM.setToken(xmWebService.TOKEN_NAME_EXCEPTION, errMsg.replace(exceptionString, ""), APXMLToken.Type.STRING);
    }

    log.debug("Exit - handleRequest");
    return responseToXM;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * handleResponse
 *
 * The main routine for handling messages from xMatters
 *
 * Handling of responses that can be dispatched based on the responseEvent token. If this function can handle the
 * response, returns true or false according to whether a response should be sent to xMatters or not.
 *
 * Responses that instead need further examination (of the response action) result in a return of undefined.
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */
function handleResponse(apxml)
{
    log.debug("Enter - handleResponse");

    // Get the additional tokens from the APXML message from xMatters, which contain the user response choice
    // and any user annotation
    var responseTokens = getAdditionalApxmlTokens(apxml);
    log.debug("handleResponse - Token keys from xMatters: [" + responseTokens + "]");

    var responseEvent = responseTokens.get("responseEvent");
    var responseAction = responseTokens.get("responseAction");
    var responder = responseTokens.get("responder");

    log.debug("handleResponse - responseEvent [" + responseEvent + "], responseAction [" + responseAction + "], responder [" + responder + "]");

    // ------------------------------------------------------------------------------------------------------------
    // The response handling can update the status of approval requests, but not change requests. However, all
    // annotations - for their own sake or relating to to changes to approval requests - are appended to the
    // relevant change request and not the approval request.
    // ------------------------------------------------------------------------------------------------------------
    var change = newRemedyObject();

    var changeRequestId = responseTokens.get("changeRequestId");
    var approvalProcessName = responseTokens.get("approvalProcessName");
    var responderId = responseTokens.get("responderId");
    var userAnnotation = responseTokens.get("annotation");

    // ------------------------------------------------------------------------------------------------------------
    // The processing of annotations is controlled by 2 parameters:
    //
    // ANNOTATE_DELIVERY (configuration.js) enables/disables the automatic annotation of delivery success/failure
    //
    // sendAnnotations token in the response enables/disables annotations (user and automatic) associated with responses.
    // ------------------------------------------------------------------------------------------------------------
    var addResponseAnnotations = equalsIgnoreCase(responseTokens.get("sendAnnotations"), "true");

    log.info("handleResponse - changeRequestId [" + changeRequestId + "], responderId [" + responderId + "], annotation [" + userAnnotation + "], addResponseAnnotations [" + addResponseAnnotations + "]");

    var xmResponseAnnotation = null;

    var changeInterfaceService = new XMCHGChangeInterfaceWS(new WSUtil(), XM_CHG_CHANGE_INTERFACE_WS_URL);
    var changeAPDetailSignatureService = new XMCHGChangeAPDetailSignatureWS(new WSUtil(), XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_URL);

    if (equalsIgnoreCase(responseEvent, "STANDALONE_ANNOTATION"))
    {
        // Don't need to retrieve the change from Remedy for this response, but need the ID set up
        change.Infrastructure_Change_ID = changeRequestId;

        // The annotation token needs to be treated as an automatic annotation from xMatters
        setChangeWorkLog(change, addResponseAnnotations, notePrefix + userAnnotation, null);
        changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, change);
    }
    else if (equalsIgnoreCase(responseEvent, "SUCCESSFUL_DELIVERY") || equalsIgnoreCase(responseEvent, "DELIVERY_FAILURE") || equalsIgnoreCase(responseEvent, "DEVICE_DELIVERED"))
    {
        if (ANNOTATE_DELIVERY)
        {
            // Don't need to retrieve the change from Remedy for this response, but need the ID set up
            change.Infrastructure_Change_ID = changeRequestId;

            xmResponseAnnotation = notePrefix + responseEvent + " to " + responder;
            setChangeWorkLog(change, true, xmResponseAnnotation, null);
            changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, change);
        }
    }
    else if (equalsIgnoreCase(responseEvent, "RECEIVED_RESPONSE"))
    {
        log.debug("handleResponse - RECEIVED_RESPONSE");

        switch (String(responseAction).toUpperCase())
        {
            case RESPONSE_ACTION_APPROVED:
            case RESPONSE_ACTION_REJECTED:
            case RESPONSE_ACTION_HOLD:
                // Construct a qualification that will return the approval request that the user is responding to.
                var qualification = XM_CHG_CHANGE_AP_DETAIL_SIGNATURE_WS_GET_APPROVAL_BY_ID_QUALIFICATION.replace("${REQUEST_ID}", String(responseTokens.get("requestId")));
                var approvalList = changeAPDetailSignatureService.getApprovals(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, qualification);

                // It is possible for a user to be listed as an approver in more than one approval for a Change Request / Process,
                // so loop over all that were found. ALL the approval records will be updated with the user's response.
                for each (var approvalListEntry in approvalList)
                {
                    // this.log.debug("approvalListEntry [" + approvalListEntry + "]");

                    // Set the approval object fields necessary to update a Remedy Approval
                    setApproval(approvalListEntry, responseAction, responderId);
                    changeAPDetailSignatureService.setApprovalStatus(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, approvalListEntry);
                }

                // An annotation of the user's response to the approval request is added to the associated change request.
                // Don't need to retrieve the change from Remedy for this response, but need the ID set up
                change.Infrastructure_Change_ID = changeRequestId;

                xmResponseAnnotation = notePrefix + "Response " + responseAction + " from " + responder;
                setChangeWorkLog(change, addResponseAnnotations, xmResponseAnnotation, userAnnotation);

                changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, change);

                break;

            case RESPONSE_ACTION_ANNOTATE:
                // Don't need to retrieve the change from Remedy for this response, but need the ID set up
                change.Infrastructure_Change_ID = changeRequestId;

                xmResponseAnnotation = notePrefix + "Change annotated by " + responder;
                setChangeWorkLog(change, addResponseAnnotations, xmResponseAnnotation, userAnnotation);

                changeInterfaceService.addWorkLog(REMEDY_WS_USERNAME, REMEDY_WS_PASSWORD, change);
                break;

            default:
                throw("Unknown responseAction [" + responseAction + "]");
                break;
        }
    }

    log.debug("handleResponse - returning change [" + change + "]");
    log.debug("Exit - handleResponse");
    return change;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * Retrieves the additional tokens from the APXML
 *
 * @param apxml XML message to process
 * @return LinkedHashMap containint the additional APXML tokens
 * @throws if the additional tokens are missing from the XML.
 * ---------------------------------------------------------------------------------------------------------------------
 */
function getAdditionalApxmlTokens(apxml)
{
    log.debug("Enter - getAdditionalApxmlTokens");

    if (isEmpty(apxml.getValue("additionaltokens")))
    {
        throw ("additionaltokens are missing from apxml");
    }

    log.debug("Exit - getAdditionalApxmlTokens");

    return xStream.fromXML(apxml.getValue("additionaltokens"));
}

